package bucash

import (
	"encoding/json"
	//"errors"
	"fmt"
	"strconv"
	"sync"
	. "template"
	"time"

	simplejson "github.com/go-simplejson"
	tml "github.com/toml-master"
)

var (
	Config       config
	CoinApi      BuCashApi
	m_unspent    []map[string]interface{}
	m_mutUnspent sync.Mutex
)

type Bucash struct {
	walletstatus bool
}

const (
	NameCoin = "BCC"
)

func (m Bucash) InitDriver(parm interface{}) bool {
	if m.initconfig() == false {
		return false
	}
	go m.getwalletinfo()

	//缓存当前未花费的
	m_unspent = make([]map[string]interface{}, 0)
	go AsyncRoutine()
	return true
}

//get the server's total available balance
func (m Bucash) CoinBalance(parm interface{}) ([]byte, error) {

	confirmedBalance, err := CoinApi.WalleGetBalance("*")
	if err != nil {
		LOG("ERROR", " GetBalance err:%v", err.Error())
		return nil, fmt.Errorf(`{"errcode":%v,"mesage":"%s"}`, GETBALANCE_ERR, getErrMsg(GETBALANCE_ERR, nil))
	}

	UnconfirmedBalance, err := CoinApi.WalleGetUnconfirmedBalance("")
	if err != nil {
		LOG("ERROR", " GetUnconfirmedBalance err:%s", err.Error())
		return nil, fmt.Errorf(`{"errcode":%v,"mesage":"%s"}`, GETUNCONFIRMEDBALANCE_ERR, getErrMsg(GETUNCONFIRMEDBALANCE_ERR, nil))
	}

	var jbuf []byte
	resp := make(map[string]interface{})
	resp["confirmedbalance"] = confirmedBalance
	resp["unconfirmed"] = UnconfirmedBalance
	resp["errcode"] = 0
	resp["exact"] = 0

	if jbuf, err = json.Marshal(resp); err != nil {
		LOG("ERROR", "lite inner json error:%v", err.Error())
		return nil, fmt.Errorf(`{"errcode":%v,"mesage":"%s"}`, INNER_ERR, getErrMsg(INNER_ERR, nil))
	}

	return jbuf, nil
}

//Returns up to 'count' most recent transactions skipping the first 'from' transactions for account 'account'.
func (m Bucash) TransInfo(parm interface{}) ([]byte, error) {

	pjs := parm.(*simplejson.Json)
	startH := pjs.Get("starth").MustUint64()
	endH := pjs.Get("endh").MustUint64()

	if startH < 1 {
		LOG("DEBUG", " input para  err:%v", startH)
		startH = 1
	}

	//get longest block count
	blockcount, err := CoinApi.WalleGetBlockCount()
	if err != nil {
		LOG("ERROR", " WalleGetBlockCount  err:%v", err.Error())
		return nil, fmt.Errorf(`{"errcode":%v,"mesage":"%s"}`, GETCLOCKCOUNT_ERR, getErrMsg(GETCLOCKCOUNT_ERR, nil))
	}

	//get startblock hash by height

	if startH > blockcount {
		LOG("DEBUG", " startheight:%v >  blockcount:%v", startH, blockcount)
		startH = blockcount
	}

	startblockhash, err := CoinApi.WalletGetBlockHash(uint64(startH - 1))
	if err != nil {
		LOG("ERROR", " WalletGetBlockHash startH:%v err:%v", startH, err.Error())
		return nil, fmt.Errorf(`{"errcode":%v,"mesage":"%s"}`, GETBLOCKINFO_ERR, getErrMsg(GETBLOCKINFO_ERR, nil))
	}

	//get endblock time by height
	if endH > blockcount {
		LOG("DEBUG", " endheight:%v >  blockcount:%v", endH, blockcount)
		endH = blockcount
	}

	endblockhash, err := CoinApi.WalletGetBlockHash(uint64(endH))
	if err != nil {
		LOG("ERROR", " WalletGetBlockHash startH:%v err:%v", startH, err.Error())
		return nil, fmt.Errorf(`{"errcode":%v,"mesage":"%s"}`, GETBLOCKINFO_ERR, getErrMsg(GETBLOCKINFO_ERR, nil))
	}

	endblocktime, err := CoinApi.WalletGetBlockTime(endblockhash)
	if err != nil {
		LOG("ERROR", " WalletGetBlockTime endH:%v err:%v", endH, err.Error())
		return nil, fmt.Errorf(`{"errcode":%v,"mesage":"%s"}`, GETBLOCKINFO_ERR, getErrMsg(GETBLOCKINFO_ERR, nil))
	}

	//Get all transactions in blocks since block [blockhash]
	resp, err := CoinApi.WalleListSinceBlock(startblockhash)
	if err != nil {
		LOG("ERROR", " WalleListSinceBlock err:%v", err.Error())
		return nil, fmt.Errorf(`{"errcode":%v,"mesage":"%s"}`, TRANSINFO_ERR, getErrMsg(TRANSINFO_ERR, nil))
	}

	var litelisttransactionsinfo []Listtransactions

	if err = json.Unmarshal([]byte(resp), &litelisttransactionsinfo); err != nil {
		LOG("ERROR", " inner json error:%v", err.Error())
		return nil, fmt.Errorf(`{"errcode":%v,"mesage":"%s"}`, INNER_ERR, getErrMsg(INNER_ERR, nil))
	}

	outputsArr := make([]map[string]interface{}, 0)

	for _, listtransaction := range litelisttransactionsinfo {

		// get all transactions from block startHright to endHeight
		if listtransaction.Blocktime <= endblocktime {
			requestData := make(map[string]interface{})
			confirmationheight, err := CoinApi.WalletGetBlockHeight(listtransaction.Blockhash)
			if err != nil {
				LOG("ERROR", "WalletGetBlockHeight error:%v", err.Error())
				continue
			}
			requestData["confirmationheight"] = confirmationheight
			requestData["address"] = listtransaction.Address
			requestData["confirmationtimestamp"] = listtransaction.Timereceived
			requestData["value"] = listtransaction.Amount
			requestData["transactionid"] = listtransaction.Txid
			requestData["confirmations"] = listtransaction.Confirmations
			requestData["walletaddress"] = true
			requestData["vout"] = listtransaction.Vout
			outputsArr = append(outputsArr, requestData)
		}
	}

	var jbuf []byte
	resptmp := make(map[string]interface{})
	resptmp["errcode"] = 0
	resptmp["cointype"] = "BCC"
	resptmp["outputs"] = outputsArr

	if jbuf, err = json.Marshal(resptmp); err != nil {
		LOG("ERROR", " inner json error:%v", err.Error())
		return nil, fmt.Errorf(`{"errcode":%v,"mesage":"%s"}`, INNER_ERR, getErrMsg(INNER_ERR, nil))
	}
	return jbuf, nil

}

//Sent an amount from an account to a lite address.
func (m Bucash) TransferAccounts(parm interface{}) ([]byte, error) {

	pjs := parm.(*simplejson.Json)
	amountTmp := pjs.Get("amount").MustString()
	destination := pjs.Get("address").MustString()

	if len(amountTmp) == 0 {
		LOG("ERROR", " amount illeagal")
		return nil, fmt.Errorf(`{"errcode":%v,"mesage":"%s"}`, AMOUNT_ERR, getErrMsg(AMOUNT_ERR, nil))
	}
	f, err := strconv.ParseFloat(amountTmp, 64)
	if err != nil {
		LOG("ERROR", " inner json error:%v", err.Error())
		return nil, fmt.Errorf(`{"errcode":%v,"mesage":"%s"}`, INNER_ERR, getErrMsg(INNER_ERR, nil))
	}

	amount := float64(f)

	transactionid, err := CoinApi.WalleSendFrom("", destination, amount, 1, "", "")
	if err != nil {
		LOG("ERROR", " Send coin error:%v", err.Error())
		return nil, fmt.Errorf(`{"errcode":%v,"mesage":"%s"}`, SENDCOINS_ERR, getErrMsg(SENDCOINS_ERR, nil))
	}

	if len(transactionid) != 0 {
		var jbuf []byte
		resp := make(map[string]interface{})
		resp["transcationids"] = []string{transactionid}
		resp["errcode"] = 0
		resp["msg"] = "ok"

		if jbuf, err = json.Marshal(resp); err != nil {
			LOG("ERROR", " inner json error:%v", err.Error())
			return nil, fmt.Errorf(`{"errcode":%v,"mesage":"%s"}`, INNER_ERR, getErrMsg(INNER_ERR, nil))
		}
		return jbuf, nil
	}
	LOG("ERROR", " transactionid is null error:%v", err.Error())
	return nil, fmt.Errorf(`{"errcode":%v,"mesage":"%s"}`, INNER_ERR, getErrMsg(INNER_ERR, nil))
}

//create a new  address&key for receiving payments
func (m Bucash) CreateAddress(parm interface{}) ([]byte, error) {

	addr, err := CoinApi.WalletNewAddr()
	if err != nil {
		LOG("ERROR", "  get new address error:%v", err.Error())
		return nil, fmt.Errorf(`{"errcode":%v,"mesage":"%s"}`, GETNEWADDRESS_ERR, getErrMsg(GETNEWADDRESS_ERR, nil))
	}
	if len(addr) != 0 {
		//get key by the address
		key, err := CoinApi.WalletDumpPrivKey(addr)

		if err != nil {
			LOG("ERROR", " dump private key error:%v", err.Error())
			return nil, fmt.Errorf(`{"errcode":%v,"mesage":"%s"}`, DUMPPRIVKEY_ERR, getErrMsg(DUMPPRIVKEY_ERR, nil))
		}
		if len(key) != 0 {

			var jbuf []byte
			resp := make(map[string]interface{})
			resp["address"] = addr
			resp["privkey"] = key
			resp["errcode"] = 0

			if jbuf, err = json.Marshal(resp); err != nil {
				LOG("ERROR", " inner json error:%v", err.Error())
				return nil, fmt.Errorf(`{"errcode":%v,"mesage":"%s"}`, INNER_ERR, getErrMsg(INNER_ERR, nil))
			}
			return jbuf, nil
		}
		LOG("ERROR", " privkey is null error:%v")
		return nil, fmt.Errorf(`{"errcode":%v,"mesage":"%s"}`, INNER_ERR, getErrMsg(INNER_ERR, nil))
	}
	LOG("ERROR", " address is null error:%v")
	return nil, fmt.Errorf(`{"errcode":%v,"mesage":"%s"}`, INNER_ERR, getErrMsg(INNER_ERR, nil))
}

func (m Bucash) getwalletinfo() bool {
	fmt.Println("WalletInfo:")
	m.walletstatus = false
	for {
		if m.walletstatus == false {
			resp, err := CoinApi.WalletInfo()
			if err != nil {
				LOG("ERROR", "WalletInfo error:%v", err.Error())
				time.Sleep(time.Second * 1)
				continue
			}

			LOG("DEBUG", "WalletInfo resp:%v", string(resp))
			m.walletstatus = true
		}

		time.Sleep(time.Second)
	}
}

//config rpc from lite.toml
func (m Bucash) initconfig() bool {
	if _, err := tml.DecodeFile("./src/bucash/bucash.toml", &Config); err != nil {
		fmt.Println(err)
		return false
	}

	fmt.Printf("%+v\n", Config)

	User := Config.Rpc.Rpcuser
	Password := Config.Rpc.Rpcpassword
	Host := Config.Rpc.Rpchost
	Port := Config.Rpc.Rpcport

	CoinApi.InitRpcConf(User, Password, Host, Port)

	initLiteErr()

	f, err := strconv.ParseFloat(Config.Limit.Fee, 64)
	if err != nil {
		LOG("ERROR", "inner json error:%v", err.Error())
		return false
	}

	fee := float64(f)
	CoinApi.WalletSetTxFee(fee)
	return true
}

func (m Bucash) ChainHeight(parm interface{}) ([]byte, error) {

	blockcount, err := CoinApi.WalleGetBlockCount()
	if err != nil {
		LOG("ERROR", "  get current block count error:%v", err.Error())
		return nil, fmt.Errorf(`{"errcode":%v,"mesage":"%s"}`, GETCLOCKCOUNT_ERR, getErrMsg(GETCLOCKCOUNT_ERR, nil))
	}

	sendstr := fmt.Sprintf(`{"errcode":%v,"height":%v}`, OK, blockcount)
	return []byte(sendstr), nil
}

func (m Bucash) ListUnspent(parm interface{}) ([]byte, error) {
	i_unspent := make([]map[string]interface{}, 0)
	pjs := parm.(*simplejson.Json)
	Issync, err := pjs.Get("Sync").Bool()
	if !Issync {
		fmt.Println("The listunspent data Asynchrony!!!!")
		i_unspent = m_unspent
	} else {
		address, err := pjs.Get("address").StringArray()
		if err != nil {
			LOG("ERROR", "  input parm error:%v", err.Error())
			return nil, fmt.Errorf(`{"errcode":%v,"mesage":"%s"}`, LISTUNSPENT_ERR, getErrMsg(LISTUNSPENT_ERR, nil))
		}
		unspent, err := CoinApi.Listunspent(address)
		if err != nil {
			LOG("ERROR", "  get listunspent:%v", err.Error())
			return nil, fmt.Errorf(`{"errcode":%v,"mesage":"%s"}`, LISTUNSPENT_ERR, getErrMsg(LISTUNSPENT_ERR, nil))
		}
		i_unspent = unspent
	}

	var jbuf []byte
	resptmp := make(map[string]interface{})
	resptmp["errcode"] = 0
	resptmp["cointype"] = "BCC"
	resptmp["unspents"] = i_unspent

	if jbuf, err = json.Marshal(resptmp); err != nil {
		LOG("ERROR", " inner json error:%v", err.Error())
		return nil, fmt.Errorf(`{"errcode":%v,"mesage":"%s"}`, INNER_ERR, getErrMsg(INNER_ERR, nil))
	}
	return jbuf, nil

}

func (m Bucash) SendRawTransaction(parm interface{}) ([]byte, error) {

	pjs := parm.(*simplejson.Json)
	transdata := pjs.Get("data").MustString()

	if len(transdata) == 0 {
		return nil, fmt.Errorf(`{"cointype":"BCC","errcode":%v,"msg":"%v"}`, PARM_ERR, getErrMsg(PARM_ERR, []byte("hexstring is null")))
	}
	transactionhex, err := CoinApi.SendRawTransaction(transdata)
	if err != nil {
		LOG("ERROR", " SendRawTransaction err:%s", err.Error())
		return nil, fmt.Errorf(`{"cointype":"BCC","errcode":%v,"msg":"%s"}`, SEND_RAW_TRANSACTION_ERR, getErrMsg(SEND_RAW_TRANSACTION_ERR, nil))
	}

	var jbuf []byte
	resp := make(map[string]interface{})
	resp["cointype"] = "BCC"
	resp["transactionids"] = transactionhex
	resp["errcode"] = 0
	resp["msg"] = "OK"

	if jbuf, err = json.Marshal(resp); err != nil {
		LOG("ERROR", "inner json error:%v", err.Error())
		return nil, fmt.Errorf(`{"cointype":"BCC","errcode":%v,"msg":"%s"}`, INNER_ERR, getErrMsg(INNER_ERR, nil))
	}
	return jbuf, nil
}

func (m Bucash) SumcoinBalance(parm interface{}) ([]byte, error) {
	return nil, nil
}
func (m Bucash) AccountsBalance(parm interface{}) (jbuf []byte, err error) {
	var balancestotal, offlinebalancetotal float64
	ilistunspent := m_unspent
	accountbalance := make([]map[string]interface{}, 0)
	offlinebalance := make([]map[string]interface{}, 0)
	for _, value := range ilistunspent {
		if value["spendable"] == true {
			balancestotal += value["Amount"].(float64)
			accountbalance = append(accountbalance, value)
		} else {
			offlinebalancetotal += value["Amount"].(float64)
			offlinebalance = append(offlinebalance, value)
		}
	}
	resptmp := make(map[string]interface{})
	resptmp["errcode"] = 0
	resptmp["cointype"] = "BCC"
	resptmp["accountbalance"] = accountbalance
	resptmp["balancestotal"] = balancestotal
	resptmp["offlinebalance"] = offlinebalance
	resptmp["offlinebalancetotal"] = offlinebalancetotal
	if jbuf, err = json.Marshal(resptmp); err != nil {
		LOG("ERROR", " inner json error:%v", err.Error())
		return nil, fmt.Errorf(`{"errcode":%v,"mesage":"%s"}`, INNER_ERR, getErrMsg(INNER_ERR, nil))
	}
	return jbuf, nil
}

func AsyncRoutine() {
	CachelistunspentTicker := time.NewTicker(120 * time.Second)
	for {
		select {
		case _ = <-CachelistunspentTicker.C:
			Cachelistunspent()
		}
	}
}
func Cachelistunspent() {
	fmt.Println("cachelistunspent time:", time.Now().Format("2006-01-02 15:04:05"))
	var address []string

	unspent, err := CoinApi.Listunspent(address)
	if err != nil {
		LOG("ERROR", "  get listunspent:%v", err.Error())
		//return nil, fmt.Errorf(`{"errcode":%v,"mesage":"%s"}`, LISTUNSPENT_ERR, getErrMsg(LISTUNSPENT_ERR, nil))
	}
	m_mutUnspent.Lock()
	m_unspent = unspent
	m_mutUnspent.Unlock()

}

func init() {
	Register(NameCoin, &Bucash{})
}
